#ifndef CROCPARTY_SYS_DATA_H
#define CROCPARTY_SYS_DATA_H

#include <stdint.h>

typedef int32_t sys_i32;
typedef uint8_t sys_color;
typedef uint32_t sys_screen_color;
typedef uint64_t sys_glyph;

sys_screen_color sys_make_screen_color(uint8_t r, uint8_t g, uint8_t b);

sys_i32 sys_max_i32(sys_i32 x, sys_i32 y); 
sys_i32 sys_min_i32(sys_i32 x, sys_i32 y); 
sys_i32 sys_abs_i32(sys_i32 x); 
sys_i32 sys_sgn_i32(sys_i32 x); 

#define SYS_COLOR_TRANSPARENT 255
#define SYS_COLOR_N 256

#endif  // CROCPARTY_SYS_DATA_H